#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <usbdrv.h>

#define ACTION_GETSTATE         1
#define ACTION_SETSTATE         2

#define LEDPIN                  PB4
#define SIGINPIN                PB0
#define SIGOUTPIN               PB3

// LittleBits signals are high when on, so we initialize our signals to off (low).
// SIGIN is pulled low by R6 but can be switched high by the previous component.
//
// Comms is done with a single byte for both read&write - bits are:
// 0: SIGIN, ie input from previous connector. 1 for on (switched to GND),
//    0 for off (pulled high)  (read-only)
// 1: SIGOUT, ie output to next connector. 1 for on (switched to GND),
//    0 for off (pulled high)  (read-write)
// 2: LED, 1 for on, 0 for off  (read-write)
// Here "previous" and "next" relate to the arrow direction on the board.
// 

volatile uint8_t signal;

int main(void) {
    uchar calibrationValue = eeprom_read_byte(0); // calibration value from last time
    if (calibrationValue != 0xff) {
        OSCCAL = calibrationValue;
    }
    usbDeviceDisconnect();
    for (int i=255;i!=0;i--) {
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    wdt_enable(WDTO_1S);
    usbInit();
    sei();

    DDRB   |= (1<<LEDPIN) | (1<<SIGOUTPIN);

    for(;;) {
        wdt_reset();
        usbPoll();
    }
}

usbMsgLen_t usbFunctionSetup(uchar data[8]) {

    usbRequest_t    *rq = (void *)data;

    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_VENDOR) {
        switch (rq->bRequest) {
        case ACTION_GETSTATE:
            signal = (signal & ~1) | (PINB&(1<<SIGINPIN) ? 1 : 0);  // Inverted
            usbMsgPtr = &signal;
            return 1;
        case ACTION_SETSTATE:
            signal = rq->wValue.word & 6;
            if (signal & 2) {               // Inverted
                PORTB &= ~(1<<SIGOUTPIN);
            } else {
                PORTB |= (1<<SIGOUTPIN);
            }
            if (signal & 4) {
                PORTB |= (1<<LEDPIN);
            } else {
                PORTB &= ~(1<<LEDPIN);
            }
            return USB_NO_MSG;
        }
    }
    return 0;
}

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
void calibrateOscillator(void) {
    uchar step = 128;
    uchar trialValue = 0, optimumValue;
    int x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    // do a binary search:
    do {
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    // proportional to current real frequency
        if (x < targetValue) {          // frequency still too low
            trialValue += step;
        }
        step >>= 1;
    } while (step > 0);
    // We have a precision of +/- 1 for optimum OSCCAL here
    // now do a neighborhood search for optimum value
    optimumValue = trialValue;
    optimumDev = x; // this is certainly far away from optimum
    for (OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++) {
        x = usbMeasureFrameLength() - targetValue;
        if (x < 0) {
            x = -x;
        }
        if (x < optimumDev) {
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}

void usbEventResetReady(void) {
    cli();
    calibrateOscillator();
    sei();
    eeprom_write_byte(0, OSCCAL);   // store the calibrated value in EEPROM
}

// Dummy USB Report Descriptor to silence drivers on Windows
PROGMEM const char usbHidReportDescriptor[22] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};
